from . import geo
from .geo import GPSPin, cleanse, split_trips


def test_parse_poly():
    thing = """who
cares
\t1.2\t3.4
\t5.6\t7.8
END 
END """
    r = geo.parse_poly(thing)
    print(f'r is {r}')
    assert len(r) == 1
    assert r[0][0] == [1.2, 3.4]


def test_parse_poly_file():
    with open('bangalore.poly') as f:
        p = geo.parse_poly(f.read())
        assert len(p) > 0


def test_cleanse():
    trajectory = open('trajectory.txt').read()
    raw = [GPSPin(lat=float(x), lng=float(y), ts=int(z))
           for x, y, z in [item.split(',') for item in trajectory.split(';')]]
    pins = []
    last = None
    for pin in raw:
        if last:
            if pin.ts == last.ts:
                continue
        last = pin
        pins.append(pin)

    r = cleanse(pins)
    assert len(r) != len(pins)


def test_split_1():
    trajectory = '1,1,0;1,2,1;1,2,2;1,2,3;1,2,4;1,2,5;1,2,6;1,2,7;1,2,8;1,3,9;1,4,10'
    pins = [GPSPin(lat=float(x), lng=float(y), ts=int(z)) for x, y, z in [
        item.split(',') for item in trajectory.split(';')]]
    trips, pauses = split_trips(pins, 100, 5)
    print(trips)
    print(pauses)
    assert len(trips) == 2
    assert len(pauses) == 1
    assert len(pauses[0]) == 8


def test_split_2():
    trajectory = '1,1,0;1,2,1;1,2,2;1,2,3;1,2,4;1,2,5;1,2,6;1,2,7;1,2,8;1,3,9;1,4,10'
    pins = [GPSPin(lat=float(x), lng=float(y), ts=int(z)) for x, y, z in [
        item.split(',') for item in trajectory.split(';')]]
    trips, pauses = split_trips(pins, 1000000000, 5)
    print(trips)
    print(pauses)
    assert len(trips) == 0
    assert len(pauses) == 1
    assert len(pauses[0]) == 11


def test_split_3():
    trajectory = '1,1,0;1,2,1;1,2,2;1,2,3;1,2,4;1,2,5;1,2,6;1,2,7;1,2,8;1,3,9;1,4,10'
    pins = [GPSPin(lat=float(x), lng=float(y), ts=int(z)) for x, y, z in [
        item.split(',') for item in trajectory.split(';')]]
    trips, pauses = split_trips(pins, 100, 12)
    print(trips)
    print(pauses)
    assert len(trips) == 1
    assert len(pauses) == 0



import math
import json
import logging
from pydantic import BaseModel


class GPSPin(BaseModel):
    lat: float
    lng: float
    ts: int

    def to_latlng(self):
        return LatLng(lat=self.lat, lng=self.lng)


"""
cleanse input pins by removing pin b if straightline distance of a-b / duration is more than speed threshold
"""


def cleanse(pins, threshold=8):
    if len(pins) < 3:
        return pins
    speed = 0
    start = pins[0]
    result = [start]
    for pin in pins[1:]:
        end = pin
        if end.ts < start.ts:
            raise ValueError('pins are not in chronological order')
        if end.ts == start.ts:
            continue
        speed = distance(LatLng(lat=start.lat, lng=start.lng), LatLng(
            lat=end.lat, lng=end.lng)) / (end.ts - start.ts)
        if speed <= threshold:
            result.append(end)
            start = end

    return result


def split_trips(pins, sld_tolerance, min_pause_duration):
    result = []
    pauses = []
    if len(pins) < 3:
        return [pins]
    start = None
    last = pins[0]
    trip = [last]
    detecting = []
    mode = 0  # 0: travelling 1: detecting pause 2: pausing
    for pin in pins[1:]:
        #print(mode, pin)
        if mode == 0:
            if distance(pin.to_latlng(), last.to_latlng()) <= sld_tolerance:
                start = last
                mode = 1
                detecting = [last, pin]
            else:
                trip.append(pin)
        elif mode == 1:
            if distance(pin.to_latlng(), start.to_latlng()) <= sld_tolerance:
                detecting.append(pin)
                if pin.ts - start.ts >= min_pause_duration:
                    mode = 2
                    if len(trip) > 1:
                        result.append(trip)
                    trip = []
            else:  # moving again
                trip.extend(detecting)
                trip.append(pin)
                detecting = []
                mode = 0
        elif mode == 2:
            if distance(pin.to_latlng(), start.to_latlng()) <= sld_tolerance:
                detecting.append(pin)
            else:  # moving again
                pauses.append(detecting)
                detecting = []
                trip = [last, pin]
                mode = 0
        last = pin
        # print(mode,len(trip),len(detecting))
    if mode == 1:  # if we end up detecting, that means no pause at all
        trip.extend(detecting)
    if len(trip) > 0:
        result.append(trip)
    if len(detecting) > 0:
        pauses.append(detecting)
    return result, pauses


class LatLng(BaseModel):
    lat: float
    lng: float


def distance(loc1: LatLng, loc2: LatLng):
    # approximate radius of earth in meter
    R = 6373000

    lat1 = math.radians(loc1.lat)
    lng1 = math.radians(loc1.lng)
    lat2 = math.radians(loc2.lat)
    lng2 = math.radians(loc2.lng)

    dlng = lng2 - lng1
    dlat = lat2 - lat1

    a = math.sin(dlat / 2)**2 + math.cos(lat1) * \
        math.cos(lat2) * math.sin(dlng / 2)**2
    c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))
    return R * c


def geojson_to_poly(gjson: str, name: str = None):
    loaded = json.loads(gjson)
    if name == None:
        if'name' in loaded:
            name = loaded['name']
        else:
            name = 'default'
    if len(loaded['features']) > 1:
        raise ValueError('multiple features geojson is not supported')
    geometry = loaded['features'][0]['geometry']
    if geometry['type'] != 'Polygon':
        raise ValueError('only converting from geojson polygon is supported')
    if len(geometry['coordinates']) > 1:
        logging.warn(
            'multiple coordinates of polygon detected, only using first one as exterior definition')
    output = [name, 'area1']
    for coord in geometry['coordinates'][0]:
        output.append(f'\t{coord[0]}\t{coord[1]}')
    output.append('END')
    output.append('END')
    return '\n'.join(output)


def poly_to_geojson(poly: str):
    gjson = {
        'type': 'FeatureCollection',
        'name': 'default',
        'features': [
            {
                'type': 'Feature',
                'properties': {},
                'geometry': {
                    'type': 'Polygon',
                    'coordinates': [
                        []
                    ]
                }
            }
        ]
    }
    poly = poly.replace('\t', ' ')
    lines = poly.split('\n')
    gjson['name'] = lines[0].strip()
    for line in lines[2:]:
        if line.strip() == 'END':
            break
        coords = [x for x in line.strip().split(' ') if len(x) > 0]
        gjson['features'][0]['geometry']['coordinates'][0].append(
            [float(coords[0]), float(coords[1])])
    return gjson


def parse_poly(pstr: str):
    result = []
    lines = pstr.split('\n')
    mode = 0
    coords = []
    for line in lines:
        line = line.rstrip()
        line = line.replace('\t', ' ')
        swt = line.startswith(' ')
        if mode == 0 and swt:
            mode = 1  # begin area
        if mode == 1 and swt:
            items = [float(x) for x in line.split(' ') if len(x) > 0]
            if len(items) == 2:
                coords.append([items[0], items[1]])
            else:
                raise ValueError(f'invalid line: {line}')
        if mode == 1 and line == 'END':
            mode = 0
            result.append(coords)
            coords = []
    return result



from typing import List, Tuple, Optional, Dict
from pydantic import BaseModel
from enum import Enum
from .geo import LatLng
from . import aiorequests
from urllib.parse import urlencode
from osrm.engine.api.fbresult import FBResult as FBI
import logging


def decode_str(obj):
    if not obj:
        return None
    if type(obj) == bytes:
        return obj.decode()
    if type(obj) == str:
        return obj


class OutputFormat(str, Enum):
    json = 'json'
    flatbuffers = 'flatbuffers'


class Geometry(str, Enum):
    polyline = 'polyline'
    polyline6 = 'polyline6'


class Code(str, Enum):
    Ok = "Ok"
    InvalidUrl = "InvalidUrl"
    InvalidService = "InvalidService"
    InvalidVersion = "InvalidVersion"
    InvalidOptions = "InvalidOptions"
    InvalidQuery = "InvalidQuery"
    InvalidValue = "InvalidValue"
    NoSegment = "NoSegment"
    TooBig = "TooBig"
    NoMatch = "NoMatch"


class Response(BaseModel):
    code: Code


LngLat = Tuple[float, float]  # format is: [lng, lat]


class WayPoint(BaseModel):
    hint: Optional[str]
    distance: float
    name: Optional[str]
    location: Optional[LngLat]
    nodes: Optional[List[int]]


class NearestResponse(Response):
    waypoints: List[WayPoint]

    @classmethod
    def from_dict(cls, dct: Dict):
        return cls(**dct)

    @classmethod
    def from_fb(cls, buf: bytes):
        fb = FBI.FBResult.GetRootAsFBResult(buf, 0)
        if fb.Error():
            code = decode_str(fb.Code().Code())
            return cls(code=code)
        code = Code.Ok
        waypoints = []
        wl = fb.WaypointsLength()
        for i in range(0, wl):
            wfb = fb.Waypoints(i)
            loc = wfb.Location()
            nodes = wfb.Nodes()
            wp = WayPoint(
                hint=decode_str(wfb.Hint()),
                location=(loc.Longtitude(), loc.Latitude()),
                name=decode_str(wfb.Name()),
                distance=wfb.Distance(),
            )
            if nodes:
                wp.nodes = [nodes.First(), nodes.Second()]
            waypoints.append(wp)

        return cls(code=code, waypoints=waypoints)


class Step(BaseModel):
    geometry: str = None


class Leg(BaseModel):
    distance: float
    duration: float
    steps: List[Step] = []


class Route(BaseModel):
    duration: Optional[float]
    distance: Optional[float]
    legs: List[Leg] = []
    geometry: str = None


class RouteResponse(Response):
    routes: List[Route] = []
    waypoints: List[WayPoint] = []

    @classmethod
    def from_dict(cls, dct: Dict):
        return cls(**dct)

    @classmethod
    def from_fb(cls, buf: bytes):
        fb = FBI.FBResult.GetRootAsFBResult(buf, 0)
        if fb.Error():
            code = decode_str(fb.Code().Code())
            return cls(code=code)
        code = Code.Ok
        routes = []
        waypoints = []
        rl = fb.RoutesLength()
        for i in range(0, rl):
            rfb = fb.Routes(i)
            route = Route(duration=rfb.Duration(), distance=rfb.Distance(), geometry=decode_str(rfb.Polyline()))
            ll = rfb.LegsLength()
            for j in range(0, ll):
                lfb = rfb.Legs(j)
                leg = Leg(duration=lfb.Duration(), distance=lfb.Distance())
                sl = lfb.StepsLength()
                for s in range(0, sl):
                    sfb = lfb.Steps(s)
                    step = Step(geometry = sfb.Geometry())
                    leg.steps.append(step)
                route.legs.append(leg)
            routes.append(route)
        return cls(code=code, routes=routes, waypoints=waypoints)


class POI(BaseModel):
    name: str
    location: LngLat


class TableResponse(Response):
    sources: List[POI] = []
    destinations: List[POI] = []
    durations: List[List[float]] = []
    distances: List[List[float]] = []

    @classmethod
    def from_dict(cls, dct: Dict):
        return cls(**dct)

    @classmethod
    def from_fb(cls, buf: bytes):
        fb = FBI.FBResult.GetRootAsFBResult(buf, 0)
        if fb.Error():
            code = decode_str(fb.Code().Code())
            return cls(code=code)
        code = Code.Ok
        tfb = fb.Table()
        durations = [[y for y in x] for x in tfb.DurationsAsNumpy()]
        distances = [[y for y in x] for x in tfb.DistancesAsNumpy()]
        return cls(code=code, duration=durations, distances=distances)


class TracePoint(BaseModel):
    matchings_index: int
    location: LngLat


class MatchResponse(Response):
    tracepoints: Optional[List[TracePoint]]
    matchings: Optional[List[Route]]

    @classmethod
    def from_dict(cls, dct: Dict):
        return cls(**dct)

    @classmethod
    def from_fb(cls, buf: bytes):
        fb = FBI.FBResult.GetRootAsFBResult(buf, 0)
        if fb.Error():
            code = decode_str(fb.Code().Code())
            return cls(code=code)
        code = Code.Ok
        matchings = []
        tracepoints = []
        rl = fb.RoutesLength()
        for i in range(0, rl):
            rfb = fb.Routes(i)
            route = Route(duration=rfb.Duration(), distance=rfb.Distance())
            matchings.append(route)
        wl = fb.WaypointsLength()
        for i in range(0, wl):
            wfb = fb.Waypoints(i)
            loc = wfb.Location()
            tp = TracePoint(matchings_index=wfb.MatchingsIndex(), location=(loc.Longtitude(), loc.Latitude()))
            tracepoints.append(tp)

        return cls(code=code, matchings=matchings, tracepoints=tracepoints)


class Gaps(str, Enum):
    """
    Allows the input track splitting based on huge timestamp gaps between points.
    """
    split = "split"
    ignore = "ignore"


def parse_google_coords(inp: str):
    return [LatLng(lat=float(t[0]), lng=float(t[1])) for t in [pair.split(',') for pair in [pairs for pairs in inp.split('|')]]]


class API:

    def __init__(self, base_url: str):
        self.base_url = base_url
        self.valid_formats = ['json', 'flatbuffers']

    async def _call(self, url: str, output_format: OutputFormat, responseCls):
        if output_format not in self.valid_formats:
            raise ValueError(f'invalid output format: {self.output_format}')
        if output_format == 'json':
            d = None
            try:
                d = await aiorequests.http_get_json(url)
            except aiorequests.CommError as ce:
                logging.error(f'err calling {url}: {ce}')
                if ce.payload:
                    return responseCls.from_dict(ce.payload)
                else:
                    raise ce
            except Exception as e:
                logging.error(f'err calling {url}: {e}')
                raise e
            try:
                return responseCls.from_dict(d)
            except Exception as e:
                logging.error(f'{responseCls} parsing {d}, {e}')
                raise e
        elif output_format == 'flatbuffers':
            return responseCls.from_fb(await aiorequests.http_get_raw(url))

    async def nearest(self, loc: LatLng, output_format: OutputFormat, number: int = 1, radius: Optional[float] = None, bearing: Optional[float] = None) -> NearestResponse:
        url = f'{self.base_url}/nearest/v1/driving/{loc.lng},{loc.lat}.{output_format}'
        params = {'number': number}
        if radius is not None:
            params['radiuses'] = radius
        if bearing is not None:
            params['bearings'] = bearing
        url += "?" + urlencode(params)
        return await self._call(url, output_format, NearestResponse)

    async def table(self, coords: List[LatLng], sources: List[int], destinations: List[int], output_format: OutputFormat) -> TableResponse:
        coords_param = ';'.join([f'{c.lng},{c.lat}' for c in coords])
        sources_param = ';'.join(sources)
        destinations_param = ';'.join(destinations)
        url = f'{self.base_url}/table/v1/driving/{coords_param}.{output_format}?sources={sources_param}&destinations={destinations_param}&annotations=duration,distance'
        return await self._call(url, output_format, TableResponse)

    async def match(self, coords: List[LatLng], output_format: OutputFormat,
                    timestamps: Optional[List[int]] = None,
                    radiuses: Optional[List[float]] = None,
                    gaps: Gaps = Gaps.split,
                    tidy: bool = False) -> MatchResponse:

        coords_param = ';'.join([f'{c.lng},{c.lat}' for c in coords])
        url = f'{self.base_url}/match/v1/driving/{coords_param}.{output_format}'
        params = {}
        if timestamps is not None:
            params['timestamps'] = ";".join(timestamps)
        if radiuses is not None:
            params['radiuses'] = ";".join(radiuses)
        params['gaps'] = gaps.value
        if tidy:
            params['tidy'] = "true"
        url += "?" + urlencode(params)
        return await self._call(url, output_format, MatchResponse)

    async def route(self, coords: List[LatLng], output_format: OutputFormat, geometry: Geometry = Geometry.polyline, steps: bool = False, alternatives: bool = False) -> RouteResponse:
        coords_param = ';'.join([f'{c.lng},{c.lat}' for c in coords])
        url = f'{self.base_url}/route/v1/driving/{coords_param}.{output_format}?overview=full&geometries={geometry}&steps={str(steps).lower()}&alternatives={str(alternatives).lower()}'
        return await self._call(url, output_format, RouteResponse)



import aiohttp
import logging

_client_session = None


class CommError(Exception):

    def __init__(self, msg, payload=None):
        super().__init__(msg)
        self.payload = payload

    pass


async def http_get_json(url, headers={}):
    return await _http_request_json(url, 'GET', None, headers)


async def http_post_json(url, payload={}, headers={}):
    return await _http_request_json(url, 'POST', payload, headers)


async def http_get_raw(url, headers={}):
    global _client_session
    if _client_session is None:
        _client_session = aiohttp.ClientSession()
    session = _client_session
    async with session.get(url, ssl=False, headers=headers) as r:
        if not (r.status >= 200 and r.status < 300):
            raise CommError(f'{url} raw get request returns {r.status}', await r.read())
        return await r.read()


async def _http_request_json(url, method, payload, headers={}):
    global _client_session
    if _client_session is None:
        _client_session = aiohttp.ClientSession()
    session = _client_session
    if method == 'GET':
        async with session.get(url, ssl=False, headers=headers) as r:
            # print(r.status,r.json())
            if not (r.status >= 200 and r.status < 300):
                raise CommError(f'{url} {method} returns {r.status}', await r.json())
            return await r.json()
    elif method == 'POST':
        async with session.post(url, ssl=False, json=payload, headers=headers) as r:
            # print(r.status,r.json())
            if not (r.status >= 200 and r.status < 300):
                raise CommError(f'{url} {method} returns {r.status}', await r.json())
            return await r.json()
    elif method == 'PATCH':
        async with session.patch(url, ssl=False, json=payload, headers=headers) as r:
            # print(r.status,r.json())
            if not (r.status >= 200 and r.status < 300):
                raise CommError(f'{url} {method} returns {r.status}', await r.json())
            return await r.json()
    elif method == 'PUT':
        async with session.put(url, ssl=False, json=payload, headers=headers) as r:
            # print(r.status,r.json())
            if not (r.status >= 200 and r.status < 300):
                raise CommError(f'{url} {method} returns {r.status}', await r.json())
            return await r.json()
    elif method == 'DELETE':
        async with session.delete(url, ssl=False, headers=headers) as r:
            if not (r.status >= 200 and r.status < 300):
                raise CommError(f'{url} {method} returns {r.status}', await r.json())
            return await r.json()



from . import model


def test_gen_trip_vector():
    schema_text = """version: 0.1
features:
  start_time:
    hour: [
        [0, 1, 2, 3, 4, 5],
        [6, 7, 8, 9, 10, 11],
        [12, 13, 14, 15, 16, 17],
        [18, 19, 20, 21, 22, 23],
      ] # hour segments: 0-6, 6-12, 12-18, 18-24
    weekday: [[1, 2, 3, 4, 5], [6, 7]] # weekday segments: Mon-Friday, Sat-Sun
  latlng: 3 # Lat/Lng precision: 3 digits after decimal point
  ETA: 60 # round(ETA / 60)
  EDA: 1000 # round(ETA / 1000)
  straight_dist: 1000 # round(straight_dist / 1000)
"""
    import datetime

    schema = model.parse_feature_schema(schema_text)
    vector = schema.gen_trip_vector(model.RawInput(
        start_time=datetime.datetime.fromisoformat(
            "2020-04-29 16:00:00+00:00").timestamp(),
        pickup_lat=1.033333,
        pickup_lng=103.111111,
        dropoff_lat=1.044444,
        dropoff_lng=103.222222,
        eda=10000,
        eta=2000,
        straight_dist=6880,
    ))

    assert vector.start_hour_class == 2
    assert vector.start_weekday_class == 0
    assert vector.pickup_lat_trunc == 1.033
    assert vector.eda_scaled == 10
    assert vector.eta_scaled == 33
    assert vector.straight_dist_scaled == 7



from .osrmapi import NearestResponse, RouteResponse, TableResponse, MatchResponse, parse_google_coords, LatLng
from .osrmapi import API
from . import aiorequests
import pytest
import asyncio


def test_parse_google_coords():
    a = parse_google_coords('1,2|3,4|5,6')
    assert len(a) == 3
    assert a[0] == LatLng(lat=1, lng=2)
    assert a[1] == LatLng(lat=3, lng=4)
    assert a[2] == LatLng(lat=5, lng=6)

    b = parse_google_coords('1,2')
    assert len(b) == 1
    assert b[0] == LatLng(lat=1, lng=2)


def test_nearest_response():
    mock = {
        'code': 'Ok',
        'waypoints': [
            {
                'hint': '',
                'distance': 1.0,
                'name': 'whocares',
                'location': [1.0, 2.0],
                'nodes':[1, 20],
            },
        ],
    }
    NearestResponse(**mock)


def test_route_response():
    mock = {
        'code': 'Ok',
        'routes': [
            {
                'geometry': 'whocares',
                'distance': 1.0,
                'duration': 1.0,
                'legs': [
                    {
                        'distance': 1.0,
                        'duration': 1.0,
                        'steps':[
                            {
                                'geometry':'whocares2',
                            }
                        ],
                    },
                ],
            },
            {
                'geometry': 'whocares',
                'distance': 2.0,
                'duration': 2.0,
                'legs': [
                    {
                        'distance': 2.0,
                        'duration': 2.0,
                        'steps':[
                            {
                                'geometry':'whocares2',
                            }
                        ],
                    },
                ],
            },
        ],
    }
    r = RouteResponse(**mock)
    assert r.routes[0].geometry == 'whocares'
    assert r.routes[0].legs[0].steps[0].geometry == 'whocares2'
    assert r.routes[1].geometry == 'whocares'
    assert r.routes[1].legs[0].steps[0].geometry == 'whocares2'


def test_table_response():
    mock = {
        'code': 'Ok',
        'sources': [
            {
                'name': 'whocares',
                'location': [1.0, 2.0],
            },
        ],
        'destinations': [
            {
                'name': 'whocares',
                'location': [1.0, 2.0],
            },
        ],
        'durations': [
            [
                1.0,
            ],
        ],
        'distances': [
            [
                1.0,
            ],
        ],
    }
    TableResponse(**mock)


def test_match_response():
    mock = {
        'code': 'Ok',
        'tracepoints': [
            {
                'matchings_index': 1,
                'location': [1.0, 2.0],
            },
        ],
        'matchings': [
            {
                'distance': 1.0,
            },
        ],
    }
    MatchResponse(**mock)


@pytest.fixture
def event_loop():
    loop = asyncio.get_event_loop()
    yield loop


async def mock_route(url):
    return {
        'code': 'Ok'
    }


def test_route(monkeypatch, event_loop):
    api = API('whocares')
    monkeypatch.setattr(aiorequests, 'http_get_json', mock_route)
    coords = [LatLng(lat=1, lng=2)]
    result = event_loop.run_until_complete(api.route(coords, 'json'))

    assert type(result) == RouteResponse

    exp = False
    try:
        event_loop.run_until_complete(api.route(coords, 'whocares'))
    except:
        exp = True
    assert exp



from typing import List, Union
from datetime import datetime
from pydantic import BaseModel
import yaml


class RawInput(BaseModel):
    start_time: float
    pickup_lat: float
    pickup_lng: float
    dropoff_lat: float
    dropoff_lng: float
    eda: float
    eta: float
    straight_dist: float


class FeatureVector(BaseModel):
    pickup_lat_trunc: float
    pickup_lng_trunc: float
    dropoff_lat_trunc: float
    dropoff_lng_trunc: float
    start_hour_class: int
    start_weekday_class: int
    eda_scaled: int
    eta_scaled: int
    straight_dist_scaled: int

    def to_list(self) -> List[Union[float, int]]:
        return [
            self.pickup_lat_trunc,
            self.pickup_lng_trunc,
            self.dropoff_lat_trunc,
            self.dropoff_lng_trunc,
            self.start_hour_class,
            self.start_weekday_class,
            self.eda_scaled,
            self.eta_scaled,
            self.straight_dist_scaled
        ]

    def field_names(self) -> List[str]:
        return [
            'pickup_lat_trunc',
            'pickup_lng_trunc',
            'dropoff_lat_trunc',
            'dropoff_lng_trunc',
            'start_hour_class',
            'start_weekday_class',
            'eda_scaled',
            'eta_scaled',
            'straight_dist_scaled'
        ]


class StartTimeEncoder:
    def __init__(self, hour_pieces=List[List[int]], weekday_pieces=List[List[int]]):
        self.hour_pieces = hour_pieces
        self.weekday_pieces = weekday_pieces

    def hour_class(self, ts: int):
        h = datetime.utcfromtimestamp(ts).hour
        # using piece index as label
        for i, p in enumerate(self.hour_pieces):
            for h2 in p:
                if h2 == h:
                    return i
        # treat all other hours not in self.hour_pieces as a label
        return len(self.hour_pieces)

    def weekday_class(self, ts: int):
        d = datetime.utcfromtimestamp(ts).isoweekday()
        # using piece index as label
        for i, p in enumerate(self.weekday_pieces):
            for d2 in p:
                if d2 == d:
                    return i
        # treat all other weekdays not in self.weekday_pieces as a label
        return len(self.hour_pieces)


class LatLngEncoder():
    def __init__(self, ndigits):
        self.ndigits = ndigits

    def truncate(self, f: float):
        return round(f, self.ndigits)


class ScaleEncoder():
    def __init__(self, unit: int):
        self.unit = unit

    def scale(self, f: float):
        return round(f / self.unit)


class FeatureSchema:
    def __init__(self, schema):
        self._check_version(schema)
        if "features" not in schema:
            raise Exception("schema.features missing")
        features = schema["features"]
        self.start_time_encoder = self._create_start_time_encoder(features)
        self.latlng_encoder = self._create_latlng_encoder(features)
        self.eta_encoder = self._create_eta_encoder(features)
        self.eda_encoder = self._create_eda_encoder(features)
        self.straight_dist_encoder = self._create_straight_dist_encoder(
            features)

    def gen_trip_vector(self, raw: RawInput) -> FeatureVector:
        return FeatureVector(
            pickup_lat_trunc=self.latlng_encoder.truncate(raw.pickup_lat),
            pickup_lng_trunc=self.latlng_encoder.truncate(raw.pickup_lng),
            dropoff_lat_trunc=self.latlng_encoder.truncate(raw.dropoff_lat),
            dropoff_lng_trunc=self.latlng_encoder.truncate(raw.dropoff_lng),
            start_hour_class=self.start_time_encoder.hour_class(
                raw.start_time),
            start_weekday_class=self.start_time_encoder.weekday_class(
                raw.start_time),
            eda_scaled=self.eda_encoder.scale(raw.eda),
            eta_scaled=self.eta_encoder.scale(raw.eta),
            straight_dist_scaled=self.straight_dist_encoder.scale(
                raw.straight_dist)
        )

    def _check_version(self, schema):
        if "version" not in schema:
            raise Exception("schema.version missing")
        # TODO: auto version number instead of hard-coded
        if str(schema["version"]) != "0.1":
            raise Exception(
                f"schema.version mismatch, expect: 0.1, got: {schema['version']}")

    def _create_start_time_encoder(self, features) -> StartTimeEncoder:
        if "start_time" not in features:
            raise Exception(
                "schema.features: start_time missing\n{}".format(features))
        if "hour" not in features["start_time"]:
            raise Exception(
                "schema.features: start_time.hour missing\n{}".format(features))
        if "weekday" not in features["start_time"]:
            raise Exception(
                "schema.features: start_time.weekday missing\n{}".format(features))
        return StartTimeEncoder(features["start_time"]["hour"], features["start_time"]["weekday"])

    def _create_latlng_encoder(self, features) -> LatLngEncoder:
        if "latlng" not in features:
            raise Exception(
                "schema.features: latlng missing\n{}".format(features))
        return LatLngEncoder(features["latlng"])

    def _create_eta_encoder(self, features) -> ScaleEncoder:
        if "ETA" not in features:
            raise Exception("feature schema: ETA missing\n{}".format(features))
        return ScaleEncoder(features["ETA"])

    def _create_eda_encoder(self, features) -> ScaleEncoder:
        if "EDA" not in features:
            raise Exception(
                "schema.features: EDA missing\n{}".format(features))
        return ScaleEncoder(features["EDA"])

    def _create_straight_dist_encoder(self, features) -> ScaleEncoder:
        if "straight_dist" not in features:
            raise Exception(
                "schema.features: straight_dist missing\n{}".format(features))
        return ScaleEncoder(features["straight_dist"])


def parse_feature_schema(text):
    schema = yaml.load(text, Loader=yaml.FullLoader)
    return FeatureSchema(schema)


if __name__ == "__main__":
    with open("./sample.feature-spec.yaml") as f:
        schema = parse_feature_schema(f.read())

    import time
    vector = schema.gen_trip_vector(RawInput(
        start_time=time.time(),
        pickup_lat=1.033333,
        pickup_lng=103.111111,
        dropoff_lat=1.044444,
        dropoff_lng=103.222222,
        eda=10000,
        eta=2000,
        straight_dist=6880,
    ))
    print(vector.field_names())
    print(vector.to_list())
